#! /usr/bin/env python3
# -*- coding: utf-8 -*-
#
# Copyright (C) 2016 Century Systems Co.,Ltd.
# Last Modified: 2016/04/25 13:39:02 kikuchi

import struct
import collections


def namedtuple_with_defaults(typename, field_names, default_values=()):
    T = collections.namedtuple(typename, field_names)
    T.__new__.__defaults__ = (None,) * len(T._fields)

    if isinstance(default_values, collections.Mapping):
        prototype = T(**default_values)
    else:
        prototype = T(*default_values)

    T.__new__.__defaults__ = tuple(prototype)
    return T


def ff(chars):
    return (chr(0xff) * chars).encode("latin1")


# [msgid: 0x01] 基本設定要求
Cmd01_set_cmd = namedtuple_with_defaults("Cmd01_set", ("type outlevel pan_id nw_id mainteip "
                                                       "mainteport aplip aplport radiusip "
                                                       "radiusport "
                                                       "ch0 ch1 ch2 ch3 ch4 "
                                                       "ch5 ch6 ch7 ch8 ch9 "
                                                       "nodetype cycle duty10 antenna "
                                                       "retry onoff rtmoni"),
                                                       {"type" : 1, "outlevel" : 3,
                                                        "mainteip" : ff(16),
                                                        "mainteport" : ff(2),
                                                        "aplip" : ff(16), "aplport" : ff(2),
                                                        "radiusip" : ff(16),
                                                        "radiusport" : ff(2),
                                                        "ch0" : 255, "ch1" : 255, "ch2" : 255,
                                                        "ch3" : 255, "ch4" : 255, "ch5" : 255,
                                                        "ch6" : 255, "ch7" : 255, "ch8" : 255,
                                                        "ch9" : 255, "nodetype" : 0xff,
                                                        "duty10" : 1800, "antenna" : 0,
                                                        "cycle" : ff(2), "retry" : 255,
                                                        "onoff" : 255, "rtmoni" : 255})

def make_cmd01_set_cmd(pan_id, nw_id, ch, outlevel=3):
    cmd01 = Cmd01_set_cmd(pan_id=pan_id, nw_id=nw_id.encode('latin1'),
                          ch0=ch, outlevel=outlevel)
    pkt = struct.pack(">BBH16s16s2s16s2s16s2s10BB2sHBBBB", *cmd01)
    return pkt

Cmd01_set_res = collections.namedtuple("Cmd01_set_res", "req type")

def decode_cmd01_set_res(pkt):
    decoded = Cmd01_set_res(*struct.unpack(">BB", pkt))
    return decoded


# [msgid: 0x01] 基本設定取得要求
Cmd01_get_cmd = namedtuple_with_defaults("Cmd01_get", "type",
                                         {"type" : 2})

def make_cmd01_get_cmd():
    cmd01 = Cmd01_get_cmd()
    pkt = struct.pack(">B", *cmd01)
    return pkt


Cmd01_get_res = collections.namedtuple("Cmd01_get_res", ("req "
                                                         "type outlevel pan_id nw_id "
                                                         "mainteip mainteport aplip aplport "
                                                         "radiusip radiusport "
                                                         "ch0 ch1 ch2 ch3 ch4 ch5 ch6 ch7 ch8 ch9 "
                                                         "nodetype cycle duty10 antenna "
                                                         "retry "
                                                         "actversion backupversion "
                                                         "onoff rtmoni"))

def decode_cmd01_get_res(pkt):
    decoded = Cmd01_get_res(*struct.unpack(">BBBH16s16s2s16s2s16s2s10BBHHBB4s4sBB", pkt))
    return decoded

# [msgid: 0x06] 局番号解析用データ設定
Cmd06_set_cmd = namedtuple_with_defaults("Cmd06_set", ("type conn_type pro rtu_bc rtu_bcadr "
                                                       "offset ascii_bcadr "
                                                       "use_stathead1 stathead1 "
                                                       "use_stathead2 stathead2 "
                                                       "adrfield adrfield_len adrformat "
                                                       "use_delimiter1 delimiter1 "
                                                       "use_delimiter2 delimiter2 "
                                                       "exp_delimiter "
                                                       "non_time other_bc other_bcadr "
                                                       "use_starthead_judge"),
                                                       {"type" : 1, "conn_type" : 0xff,
                                                        "pro" : 1, "rtu_bc" : 0,
                                                        "rtu_bcadr" : 0, "offset" : 0,
                                                        "ascii_bcadr" : 0,
                                                        "use_stathead1" : 0, "stathead1" : 0,
                                                        "use_stathead2" : 0, "stathead2" : 0,
                                                        "adrfield" : 0, "adrfield_len" : 0,
                                                        "adrformat" : 0,
                                                        "use_delimiter1" : 0, "delimiter1" : 0,
                                                        "use_delimiter2" : 0, "delimiter2" : 0,
                                                        "exp_delimiter" : 0, "non_time" : 0,
                                                        "other_bc" : 0, "other_bcadr" : 0,
                                                       "use_starthead_judge" : 0})

def make_cmd06_set_cmd():
    cmd06 = Cmd06_set_cmd()
    pkt = struct.pack(">BBBBBBHBBBBBBBBBBBBBBLB", *cmd06)
    return pkt

Cmd06_set_res = collections.namedtuple("Cmd06_set_res", "req type")

def decode_cmd06_set_res(pkt):
    decoded = Cmd06_set_res(*struct.unpack(">BB", pkt))
    return decoded

# [msgid: 0x08] 経路情報取得要求
Cmd08_get_cmd = namedtuple_with_defaults("Cmd08_get", "type mac",
                                         {"type" : 1, "mac": ff(16)})

def make_cmd08_get_cmd(typ=2, mac=None):
    if not mac:
        mac = ff(16)
    elif len(mac) != 8:
        raise
    elif isinstance(mac, str):
        mac = mac.encode('latin1')

    if typ == 1:
        str_unpack_fmt = ">B"
        Cmd08_get_cmd = namedtuple_with_defaults("Cmd08_get", "type")
        cmd08 = Cmd08_get_cmd(type=typ)
    elif typ == 2:
        str_unpack_fmt = ">B8s"
        Cmd08_get_cmd = namedtuple_with_defaults("Cmd08_get", "type mac")
        cmd08 = Cmd08_get_cmd(type=typ, mac=mac)
    else:
        raise

    pkt = struct.pack(str_unpack_fmt, *cmd08)
    return pkt

def decode_cmd08_get_res(pkt):
    if len(pkt) == 1:
        str_field_names = "req"
        str_unpack_fmt = ">B"
    elif len(pkt) == 21:
        str_field_names = "req end num total dodag"
        str_unpack_fmt = ">BBBH16s"
    elif (len(pkt)-21) % 24 == 0:
        # TODO num > 1
        str_field_names = "req end num total dodag global_ parent mac parentmac"
        str_unpack_fmt = ">BBBH16s16s16s8s8s"
    else:
        raise

    Cmd08_get_res = collections.namedtuple("Cmd08_get_res", (str_field_names))
    decoded = Cmd08_get_res(*struct.unpack(str_unpack_fmt, pkt))
    return decoded

# [msgid: 0x0C] リセット要求
Cmd0C_cmd = namedtuple_with_defaults("Cmd09", "sec boot",
                                     {"sec" : 0, "boot" : 0})

def make_cmd0c_cmd(sec=0, boot=0):
    cmd0c = Cmd0C_cmd(sec=sec, boot=boot)
    pkt = struct.pack(">HB", *cmd0c)
    return pkt

# [msgid: 0x10] 暗号化鍵登録要求
Cmd10_set_cmd = namedtuple_with_defaults("Cmd10_set", "nwkey apikey",
                                         {"nwkey" : ff(32), "apikey" : ff(32)})

def make_cmd10_set_cmd(nwkey=None):
    if not nwkey:
        nwkey = ff(32)
    elif len(nwkey) != 16:
        raise
    elif isinstance(nwkey, str):
        nwkey = nwkey.encode('latin1')

    cmd10 = Cmd10_set_cmd(nwkey=nwkey)
    pkt = struct.pack(">32s32s", *cmd10)
    return pkt

Cmd10_set_res = collections.namedtuple("Cmd10_set_res", ("req "
                                                         "nwkey aplkey"))

def decode_cmd10_set_res(pkt):
    decoded = Cmd10_set_res(*struct.unpack(">B32s32s", pkt))
    return decoded

# [msgid: 0x1A] モジュール状態情報取得要求
Cmd1A_set_cmd = namedtuple_with_defaults("Cmd1A_set", "type",
                                         {"type" : ff(1)})

def make_cmd1A_set_cmd(typ=1):
    cmd1a = Cmd1A_set_cmd(type=typ)
    pkt = struct.pack(">B", *cmd1a)
    return pkt

Cmd1A_set_res = collections.namedtuple("Cmd1A_set_res", ("req type ipaddr "
                                                         "macidr panid nwid "
                                                         "dodag ch actled "
                                                         "rssiled lonlyled "
                                                         "time"))

def decode_cmd1A_set_res(pkt):
    if len(pkt) != 66:
        return None
    decoded = Cmd1A_set_res(*struct.unpack(">BB16s8sH16s16sBBBBH", pkt))
    return decoded

# [msgid: 0x21] ノード生存確認要求
Cmd21_send_cmd = namedtuple_with_defaults("Cmd21_set", "ipv6adr",
                                         {"ipv6adr" : ff(32)})

def make_cmd21_send_cmd(ipaddress=None):
    if not ipaddress:
        ipaddress = ff(32)
    elif len(ipaddress) != 16:
        raise
    elif isinstance(ipaddress, str):
        ipaddress = ipaddress.encode('latin1')

    cmd21 = Cmd21_send_cmd(ipv6adr=ipaddress)
    pkt = struct.pack(">16s", *cmd21)
    return pkt

def decode_cmd21_send_res(pkt):
    if len(pkt) == 17:
        str_field_names = "req ipv6addr"
        str_unpack_fmt = ">B16s"
    elif len(pkt) == 1:
        str_field_names = "req"
        str_unpack_fmt = ">B"
    else:
        raise

    Cmd21_get_res = collections.namedtuple("Cmd21_get_res", (str_field_names))
    decoded = Cmd21_get_res(*struct.unpack(str_unpack_fmt, pkt))
    return decoded

# [msgid: 0x23] パケットフィルタリング機能設定
Cmd23_fields = "type Pfilter Timeout Resptime " + \
  ' '.join(["Rsv{}".format(x) for x in range(1, 14)])
Cmd23_defaults = {"type" : 1}
Cmd23_defaults.update({"Rsv{}".format(x) : 255 for x in range(1, 14)})

Cmd23_set_cmd = namedtuple_with_defaults("Cmd23_set", Cmd23_fields, Cmd23_defaults)

def make_cmd23_set_cmd(Pfilter, Timeout, Resptime):
    cmd23 = Cmd23_set_cmd(Pfilter=int(Pfilter), Timeout=Timeout, Resptime=Resptime)
    pkt = struct.pack(">BBHH13B", *cmd23)
    return pkt

Cmd23_set_res = collections.namedtuple("Cmd23_set_res", "req type")

def decode_cmd23_set_res(pkt):
    decoded = Cmd23_set_res(*struct.unpack(">BB", pkt))
    return decoded

# [msgid: 0x88] 局番号リスト設定
Cmd88_set_cmd = namedtuple_with_defaults("Cmd88_set", "type pro adrfield adrformat page",
                                         {"type" : 1, "pro" : 1, "adrfield" : 0,
                                          "adrformat" : 0})

cmd88_addr_struct = struct.Struct(">LH")

def make_cmd88_cmd(page, addr_table):
    if isinstance(addr_table, list) != True:
        return None

    if len(addr_table) > 150:
        return None

    cmd88 = Cmd88_set_cmd(page=page)
    pkt_first = struct.pack(">BBBBB", *cmd88)

    l = []
    for tpl in addr_table:
        if isinstance(tpl, tuple) and len(tpl) == 2:
            try:
                l.append(cmd88_addr_struct.pack(*tpl))
            except:
                pass

    for i in range(150 - len(l)):
        l.append(b"\x00" * 6)

    pkt_last = b''.join(l)
    pkt = pkt_first + pkt_last

    return pkt

Cmd88_set_res = collections.namedtuple("Cmd88_set_res", "req type")

def decode_cmd88_set_res(pkt):
    decoded = Cmd88_set_res(*struct.unpack(">BB", pkt))
    return decoded

# [msgid: 0xD3] 運用開始要求
CmdD3_set_res = collections.namedtuple("CmdD3_set_res", "req")

def decode_cmdD3_set_res(pkt):
    decoded = CmdD3_set_res(*struct.unpack(">B", pkt))
    return decoded


# buffer -> frame

def make_frame(msgid, seq, data):
    pass
